# Examples
## Profit Maximization Example: using Nonconvex.jl
This a example on how to use [Nonconvex.jl](https://github.com/JuliaNonconvex/Nonconvex.jl) to maximize the profit of a company operating in a defined market (through the `OptimalBids` API).

 - For this example, we will use a market of type `PowerModelsMarkets` which represents an energy spot market.
 - Case instance is IEEE 118 Bus Case
 - To ilustrate a situation where the company defines the distribution of volume across it's assets in a predefined way (and reduce the decision dimentionality), the only controlable variable to maximize the company profit is a multiplicative factor of all offers.  

### Include Needed Packages

```@example Nonconvex
using OptimalBids
using OptimalBids.PowerModelsMarkets
using Clp # Market Clearing Solver
using JuMP: optimizer_with_attributes

using Nonconvex
using NonconvexIpopt # Nonconvex.@load Ipopt
using NonconvexBayesian # Nonconvex.@load BayesOpt

using Plots # For some evaluation plots at the end

```

### Case Definition

```@example Nonconvex

# Read market data from IEEE 118 bus case
case_name = "case118.m"
DATA_DIR = joinpath(dirname(dirname(dirname(@__DIR__))), "test/data")
case_file_path = joinpath(DATA_DIR, case_name)
network_data = PowerModels.parse_file(case_file_path)

# Let's make the case a bit more interesting, by adding some randomness to existing generators costs and available load.
using Random
Random.seed!(654654)
for gen in values(network_data["gen"])
    gen["cost"][end-1] += rand(-2000.0:2500.0)
end
load_mul_factor = 6.8
for load in collect(values(network_data["load"]))[1:20:end]
    load["pd"] *= load_mul_factor * rand(0.01:0.01:3)
end

# Pretend we are a company constructing a new set of generators in the grid.
# Choose a percentage of the total number of buses to install the new generators:
percentage_buses = 0.09

# We need the keys PowerModels uses to reference the appropriate buses in it's network data dictionary.
# First, find out all available keys:
bus_indexes = collect(keys(network_data["bus"]))
# Then, calculate number of buses that consitute the chose percent (`percentage_buses`):
num_buses = length(bus_indexes)
num_strategic_buses = ceil(Int, percentage_buses * num_buses)
# To avoid any biases let's grab some generators in the middle:
bus_indexes = bus_indexes[21:(21 + num_strategic_buses - 1)]
# Finally, add new generators to the network grid data and collect their reference keys.
generator_indexes = [
    add_generator(network_data, parse(Int, bus_idx)) for bus_idx in bus_indexes
]

```

### OptimalBids API

```@example Nonconvex

# Define market
market = build_market(
    PowerModelsMarket,
    network_data,
    generator_indexes,
    optimizer_with_attributes(Clp.Optimizer, "LogLevel" => 0),
)

# Relative distribution of offers are sometimes predefined and cannot be changed at bidding time.
offer_weights = rand(num_strategic_buses)
offer_weights = offer_weights/ sum(offer_weights)

# However, the decision maker is allowed to increase all bids evenly:
min_total_volume = 0.0
max_total_volume = 65.0
range_mul_factor = min_total_volume:0.1:max_total_volume
bid_range = [offer_weights .* [i] for i in range_mul_factor]
p_curve = profit_curve!(market, bid_range)

# Let's plot and see how the range profit evaluatiuon went:
plt_range = plot(collect(range_mul_factor), p_curve,
    label="Range Evaluation",
    ylabel="Profit (\$)",
    xlabel="Multiplicative Factor",
    legend=:outertopright,
);
plt_comp = deepcopy(plt_range);
```

### Nonconvex API

```@example Nonconvex

# Nonconvex needs a minimization objective function that only receives the decision vector.
function profit_function(total_volume)
    global fcalls += 1
    return - profit_for_bid!(market, offer_weights .* total_volume[1])
end

```

### BayesOpt (0-Order)

```@example Nonconvex

# Max Number of Iterations for the solution method (proxy to a time limit at bidding time).
# ps.: Currently, no option for limiting fcalls.
maxiter = 10
global fcalls = 0

# Build Nonconvex optimization model:
model = Nonconvex.Model()
set_objective!(model, profit_function, flags = [:expensive])
addvar!(model, [min_total_volume], [max_total_volume])
add_ineq_constraint!(model, x -> -1) # errors when no inequality is added!

# Solution Method: Bayesian Optimization
alg = BayesOptAlg(IpoptAlg())
options = BayesOptOptions(
    sub_options = IpoptOptions(max_iter = maxiter),
    maxiter = maxiter, ftol = 1e-4, ctol = 1e-5, initialize=false,
)

# Optimize model:
r_bayes = optimize(model, alg, [min_total_volume]; options = options)

best_solution = r_bayes.minimizer
best_profit = -r_bayes.minimum

scatter!(plt_comp, [best_solution; r_bayes.sub_result.minimizer], [best_profit; -r_bayes.sub_result.minimum],
    label="BayesOpt - OPF Calls:$(fcalls)",
)

plt_surrogate = deepcopy(plt_range)

plot!(plt_surrogate, range_mul_factor, -getproperty.(r_bayes.surrogates[1].(range_mul_factor), :lo),
    title="BayesOpt Analysis",
    label="Surrogate Function",
)
```

### NLopt - BOBYQA (0-Order)

```@example Nonconvex

using NonconvexNLopt

maxeval = 4
global fcalls = 0

# Build Nonconvex optimization model:
model = Nonconvex.Model()
set_objective!(model, profit_function)
addvar!(model, [min_total_volume], [max_total_volume])

# Solution Method: Sequential Least-Squares Quadratic Programming
method = :LN_BOBYQA
alg = NLoptAlg(:LN_BOBYQA)
options = NLoptOptions(maxeval=maxeval)

# Optimize model
r = optimize(model, alg, [min_total_volume], options = options)

best_solution = r.minimizer
best_profit = -r.minimum

scatter!(plt_comp, [best_solution], [best_profit],
    label="NLOpt-$(method) - OPF Calls:$(fcalls)",
);

println("OPF Evaluations: ", fcalls)
```

### NonconvexMultistart - GPSampler (0-Order)

```@example Nonconvex

using NonconvexMultistart

# Build Nonconvex optimization model:
model = Nonconvex.Model()
set_objective!(model, profit_function)
addvar!(model, [min_total_volume], [max_total_volume])

# Solution Method: Hyperopt
global fcalls = 0
maxiter = 4
method = :Hyperopt
alg = HyperoptAlg(IpoptAlg())
options = HyperoptOptions(
    sub_options = IpoptOptions(max_iter = maxiter), sampler = GPSampler(),
    iters = 2,
    keep_all=true
)

# Optimize model
r_hyp = optimize(model, alg, [min_total_volume], options = options)

best_solution = r_hyp.minimizer
best_profit = -r_hyp.minimum

scatter!(plt_comp, best_solution, [best_profit],
    label="$(method) Offer - OPF Calls:$(fcalls)",
);

plt_visited = deepcopy(plt_range)

scatter!(plt_visited, [i.minimizer[1] for i in r_hyp.results], [- i.minimum for i in r_hyp.results],
    label="$(method) Offer - OPF Calls:$(fcalls)",
    title="Hyperopt Visited Offers",
)
```

### Profit Comparison NLP 0-Order Strategies

```@example Nonconvex

plot(plt_comp, margin=5Plots.mm,
    title="Profit Comparison NLP Strategies",
)
```

### NLopt - CCSAQ (1-Order)

```@example Nonconvex

using NonconvexNLopt

maxeval = 4
global fcalls = 0

# Build Nonconvex optimization model:
model = Nonconvex.Model()
set_objective!(model, profit_function)
addvar!(model, [min_total_volume], [max_total_volume])

# Solution Method: Sequential Least-Squares Quadratic Programming
method = :LD_CCSAQ
alg = NLoptAlg(method)
options = NLoptOptions(maxeval=maxeval)

# Optimize model
r = optimize(model, alg, [min_total_volume], options = options)

best_solution = r.minimizer
best_profit = -r.minimum

scatter!(plt_comp, [best_solution], [best_profit],
    label="NLOpt-$(method) - OPF Calls:$(fcalls)",
)
```